# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'term_deposits_report.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QDateEdit, QFrame,
    QGridLayout, QHeaderView, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

from jal.widgets.custom.tableview_with_footer import TableViewWithFooter

class Ui_TermDepositsReportWidget(object):
    def setupUi(self, TermDepositsReportWidget):
        if not TermDepositsReportWidget.objectName():
            TermDepositsReportWidget.setObjectName(u"TermDepositsReportWidget")
        TermDepositsReportWidget.resize(769, 338)
        self.verticalLayout = QVBoxLayout(TermDepositsReportWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.ReportParamsFrame = QFrame(TermDepositsReportWidget)
        self.ReportParamsFrame.setObjectName(u"ReportParamsFrame")
        self.ReportParamsFrame.setFrameShape(QFrame.Panel)
        self.ReportParamsFrame.setFrameShadow(QFrame.Sunken)
        self.gridLayout = QGridLayout(self.ReportParamsFrame)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(2, 2, 2, 2)
        self.ReportFrameSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.ReportFrameSpacer, 0, 1, 1, 1)

        self.SaveButton = QPushButton(self.ReportParamsFrame)
        self.SaveButton.setObjectName(u"SaveButton")

        self.gridLayout.addWidget(self.SaveButton, 0, 2, 1, 1)

        self.DepositsDate = QDateEdit(self.ReportParamsFrame)
        self.DepositsDate.setObjectName(u"DepositsDate")
        self.DepositsDate.setDateTime(QDateTime(QDate(2020, 11, 24), QTime(0, 0, 0)))
        self.DepositsDate.setCalendarPopup(True)
        self.DepositsDate.setTimeSpec(Qt.UTC)

        self.gridLayout.addWidget(self.DepositsDate, 0, 0, 1, 1)


        self.verticalLayout.addWidget(self.ReportParamsFrame)

        self.ReportTableView = TableViewWithFooter(TermDepositsReportWidget)
        self.ReportTableView.setObjectName(u"ReportTableView")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(4)
        sizePolicy.setHeightForWidth(self.ReportTableView.sizePolicy().hasHeightForWidth())
        self.ReportTableView.setSizePolicy(sizePolicy)
        self.ReportTableView.setFrameShape(QFrame.Panel)
        self.ReportTableView.setFrameShadow(QFrame.Sunken)
        self.ReportTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.ReportTableView.setAlternatingRowColors(True)
        self.ReportTableView.setGridStyle(Qt.DotLine)
        self.ReportTableView.setWordWrap(False)
        self.ReportTableView.verticalHeader().setVisible(False)
        self.ReportTableView.verticalHeader().setMinimumSectionSize(20)
        self.ReportTableView.verticalHeader().setDefaultSectionSize(20)

        self.verticalLayout.addWidget(self.ReportTableView)


        self.retranslateUi(TermDepositsReportWidget)

        QMetaObject.connectSlotsByName(TermDepositsReportWidget)
    # setupUi

    def retranslateUi(self, TermDepositsReportWidget):
        TermDepositsReportWidget.setWindowTitle(QCoreApplication.translate("TermDepositsReportWidget", u"Term deposits report", None))
        self.SaveButton.setText(QCoreApplication.translate("TermDepositsReportWidget", u"Save...", None))
        self.DepositsDate.setDisplayFormat(QCoreApplication.translate("TermDepositsReportWidget", u"dd/MM/yyyy", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'login_lidl_plus_dlg.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWidgets import (QApplication, QDialog, QFrame, QPushButton,
    QSizePolicy, QVBoxLayout, QWidget)

class Ui_LoginLidlPlusDialog(object):
    def setupUi(self, LoginLidlPlusDialog):
        if not LoginLidlPlusDialog.objectName():
            LoginLidlPlusDialog.setObjectName(u"LoginLidlPlusDialog")
        LoginLidlPlusDialog.resize(400, 500)
        self.verticalLayout_3 = QVBoxLayout(LoginLidlPlusDialog)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(2, 2, 2, 2)
        self.LidlPlusWebView = QWebEngineView(LoginLidlPlusDialog)
        self.LidlPlusWebView.setObjectName(u"LidlPlusWebView")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.LidlPlusWebView.sizePolicy().hasHeightForWidth())
        self.LidlPlusWebView.setSizePolicy(sizePolicy)
        self.LidlPlusWebView.setUrl(QUrl(u"about:blank"))

        self.verticalLayout_3.addWidget(self.LidlPlusWebView)

        self.ButtonFrame = QFrame(LoginLidlPlusDialog)
        self.ButtonFrame.setObjectName(u"ButtonFrame")
        self.ButtonFrame.setFrameShape(QFrame.NoFrame)
        self.ButtonFrame.setFrameShadow(QFrame.Plain)
        self.verticalLayout_6 = QVBoxLayout(self.ButtonFrame)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 6)
        self.SplitLine = QFrame(self.ButtonFrame)
        self.SplitLine.setObjectName(u"SplitLine")
        self.SplitLine.setFrameShape(QFrame.HLine)
        self.SplitLine.setFrameShadow(QFrame.Sunken)

        self.verticalLayout_6.addWidget(self.SplitLine)

        self.CloseBtn = QPushButton(self.ButtonFrame)
        self.CloseBtn.setObjectName(u"CloseBtn")
        sizePolicy1 = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.CloseBtn.sizePolicy().hasHeightForWidth())
        self.CloseBtn.setSizePolicy(sizePolicy1)

        self.verticalLayout_6.addWidget(self.CloseBtn, 0, Qt.AlignHCenter)


        self.verticalLayout_3.addWidget(self.ButtonFrame)


        self.retranslateUi(LoginLidlPlusDialog)
        self.CloseBtn.clicked.connect(LoginLidlPlusDialog.close)

        QMetaObject.connectSlotsByName(LoginLidlPlusDialog)
    # setupUi

    def retranslateUi(self, LoginLidlPlusDialog):
        LoginLidlPlusDialog.setWindowTitle(QCoreApplication.translate("LoginLidlPlusDialog", u"Authorization Lidl Plus", None))
        self.CloseBtn.setText(QCoreApplication.translate("LoginLidlPlusDialog", u"Close", None))
    # retranslateUi




from decimal import Decimal
from PySide6.QtCore import Qt, Slot, QDate, QAbstractTableModel, QModelIndex
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QStyledItemDelegate, QHeaderView
from jal.constants import CustomColor, Setup
from jal.db.ledger import Ledger
from jal.db.helpers import localize_decimal
from jal.db.operations import LedgerTransaction
from jal.widgets.helpers import ts2dt


#-----------------------------------------------------------------------------------------------------------------------
# Function returns True if given number has move decimal places than Setup.DEFAULT_ACCOUNT_PRECISION
def long_fraction(x: Decimal) -> bool:
    if x is None:
        return False
    if x.is_nan():
        return False
    return abs(x - round(x, Setup.DEFAULT_ACCOUNT_PRECISION)) > Decimal('0')
#-----------------------------------------------------------------------------------------------------------------------

class OperationsModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [" ", self.tr("Timestamp"), self.tr("Account"), self.tr("Notes"),
                         self.tr("Amount"), self.tr("Balance"), self.tr("Currency")]
        self._view = parent_view
        self._amount_delegate = None
        self._total_delegate = None
        self._data = []
        self._begin = 0
        self._end = 0
        self._account = 0
        self._bold_font = QFontDatabase.systemFont(QFontDatabase.GeneralFont)
        self._bold_font.setBold(True)

        self.prepareData()

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def get_operation(self, row):
        if (row >= 0) and (row < self.rowCount()):
            return self._data[row]['op_type'], self._data[row]['id']
        else:
            return [0, 0]

    def data(self, index, role=Qt.DisplayRole, field=''):
        row = index.row()
        if not index.isValid():
            return None
        try:
            operation = LedgerTransaction().get_operation(self._data[row]['op_type'],
                                                          self._data[row]['id'],
                                                          self._data[row]['subtype'])
        except IndexError as e:
            if str(e) == LedgerTransaction.NoOpException:
                return None
            raise e
        if role == Qt.DisplayRole:
            return self.data_text(operation, index.column())
        if role == Qt.FontRole and index.column() == 0:
            # below line isn't related with font, it is put here to be called for each row minimal times (ideally 1)
            self._view.setRowHeight(row, self._view.verticalHeader().fontMetrics().height() * operation.view_rows())
            return self._bold_font
        if role == Qt.ForegroundRole and self._view.isEnabled():
            if index.column() == 0:
                return operation.label_color()
            if index.column() == 5 and operation.reconciled():
                return CustomColor.Blue
        if role == Qt.ToolTipRole:
            if index.column() == 0:
                return operation.name()
            elif index.column() == 4 or index.column() == 5:
                data = self.data_text(operation, index.column())
                if any([long_fraction(x) for x in data]):
                    return '\n'.join([localize_decimal(x) for x in data])
        if role == Qt.TextAlignmentRole:
            if index.column() == 0:
                return int(Qt.AlignCenter | Qt.AlignVCenter)
            if index.column() == 4 or index.column() == 5:
                return int(Qt.AlignRight)
            return int(Qt.AlignLeft)
        if role == Qt.UserRole:  # return underlying data for given field extra parameter
            return self._data[index.row()][field]

    def data_text(self, operation, column):
        if column == 0:
            return operation.label()
        elif column == 1:
            date_time = ts2dt(operation.timestamp())
            if operation.number()  and operation.type() != LedgerTransaction.Transfer:  # Transfer is 1-liner
                date_time += f"\n# {operation.number()}"
            return date_time
        elif column == 2:
            if operation.asset_name() and operation.type() != LedgerTransaction.Transfer:
                return operation.account_name() + "\n" + operation.asset_name()
            else:
                return operation.account_name()
        elif column == 3:
            return operation.description()
        elif column == 4:
            return operation.value_change()
        elif column == 5:
            return operation.value_total()
        elif column == 6:
            return operation.value_currency()
        else:
            assert False, "Unexpected column number"

    def configureView(self):
        self._view.setColumnWidth(0, 10)
        self._view.setColumnWidth(1, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        self._view.setColumnWidth(2, 300)
        self._view.horizontalHeader().setSectionResizeMode(3, QHeaderView.Stretch)
        self._view.horizontalHeader().setFont(self._bold_font)
        self._amount_delegate = ColoredAmountsDelegate(self._view)
        self._total_delegate = ColoredAmountsDelegate(self._view, colors=False, signs=False)
        self._view.setItemDelegateForColumn(4, self._amount_delegate)
        self._view.setItemDelegateForColumn(5, self._total_delegate)

        self._view.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)  # row size is adjusted in data() method

    @Slot()
    def setAccount(self, account_id):
        if self._account != account_id:
            self._account = account_id
            self.prepareData()

    def getAccount(self):
        return self._account

    @Slot()
    def setDateRange(self, start, end=0):
        self._begin = start
        if end:
            self._end = end
        else:
            self._end = QDate.currentDate().endOfDay(Qt.UTC).toSecsSinceEpoch()
        self.prepareData()

    def update(self):
        self.prepareData()

    @Slot()
    def refresh(self):
        idx = self._view.selectionModel().selection().indexes()
        idx = idx[0] if idx else QModelIndex()  # Take first selected or empty index
        if self._view.model() != self:          # View uses some proxy model
            idx = self._view.model().mapToSource(idx)
        self.prepareData()
        if self._view.model() != self:
            idx = self._view.model().mapFromSource(idx)
        self._view.setCurrentIndex(idx)

    def prepareData(self):
        self._data = []
        self._data = Ledger.get_operations_sequence(self._begin, self._end, self._account)
        self.modelReset.emit()

    def delete_rows(self, rows):
        for row in rows:
            if (row >= 0) and (row < len(self._data)):
                LedgerTransaction.get_operation(self._data[row]['op_type'], self._data[row]['id'],
                                                display_type=self._data[row]['subtype']).delete()
        self.prepareData()

    def assign_tag_to_rows(self, rows, tag_id):
        for row in rows:
            if (row >= 0) and (row < len(self._data)):
                LedgerTransaction.get_operation(self._data[row]['op_type'], self._data[row]['id'],
                                                display_type=self._data[row]['subtype']).assign_tag(tag_id)
        self.prepareData()


# ----------------------------------------------------------------------------------------------------------------------
# Display several numbers that provided by the model in form of a list
# Each number is displayed on its own line
# colors - display positive/negative values with green/red color
# signs - display +/- sign before the number if True or only "-" if False
class ColoredAmountsDelegate(QStyledItemDelegate):
    def __init__(self, parent=None, colors=True, signs=True):
        self._view = parent
        self._colors = colors
        self._signs = signs
        super().__init__(parent=parent)

    def paint(self, painter, option, index):
        data = index.model().data(index)
        if not data:
            return
        painter.save()
        color = index.model().data(index, role = Qt.ForegroundRole)
        rect = option.rect
        H = rect.height()
        Y = rect.top()
        rect.setHeight(H / len(data))
        for i, item in enumerate(data):
            rect.moveTop(Y + i * (H / len(data)))
            self.draw_value(option.rect, painter, item, color)
        painter.restore()

    # Displays given value as formatted number with required color (or Green/Red if self._colors is True)
    # If value is None - displays nothing, If value is NaN - displays Setup.NULL_VALUE
    def draw_value(self, rect, painter, value, color=None):
        text = localize_decimal(value, precision=2, sign=self._signs)
        pen = painter.pen()
        try:
            if self._view.isEnabled():
                if self._colors:
                    if value >= 0:
                        pen.setColor(CustomColor.DarkGreen)
                    else:
                        pen.setColor(CustomColor.DarkRed)
                else:
                    if color is not None:
                        pen.setColor(color)
            painter.setPen(pen)
            painter.drawText(rect, Qt.AlignRight | Qt.AlignVCenter, text)
            if long_fraction(value):  # Underline decimal part
                shift = painter.fontMetrics().horizontalAdvance(text[-Setup.DEFAULT_ACCOUNT_PRECISION:])
                painter.drawLine(rect.right() - shift, rect.bottom(), rect.right(), rect.bottom())
        except TypeError:
            pass



from datetime import datetime, timezone
import logging
import pandas
from jal.data_import.statement import FOF, Statement, Statement_Capabilities
from jal.net.downloader import QuoteDownloader

JAL_STATEMENT_CLASS = "StatementRevolutCrypto"


# ----------------------------------------------------------------------------------------------------------------------
class StatementRevolutCrypto(Statement):
    ACCOUNT_ID = 1

    def __init__(self):
        super().__init__()
        self.name = self.tr("Revolut / Crypto")
        self.icon_name = "revolut.png"
        self.filename_filter = self.tr("Revolut statement (*.csv)")
        self._statement = None
        self.coinbase_assets = []

    @staticmethod
    def capabilities() -> set:
        return {Statement_Capabilities.MULTIPLE_LOAD}

    @staticmethod
    def order_statements(statementFiles) -> list:
        return statementFiles   # We don't care about order of revolut statements loading as they are per asset

    def load(self, filename: str) -> None:
        self._data = {
            FOF.PERIOD: [int(datetime.now(tz=timezone.utc).timestamp())] * 2,  # Set the latest timestamp to prevent warning message
            FOF.ACCOUNTS: [],
            FOF.ASSETS: [],
            FOF.SYMBOLS: [],
            FOF.ASSETS_DATA: [],
            FOF.TRADES: [],
            FOF.TRANSFERS: [],
            FOF.CORP_ACTIONS: [],
            FOF.ASSET_PAYMENTS: [],
            FOF.INCOME_SPENDING: []
        }

        self._statement = pandas.read_csv(filename, header=0)
        self._statement['Started Date'] = pandas.to_datetime(self._statement['Started Date'], format='%Y-%m-%d %H:%M:%S')
        self._statement['Completed Date'] = pandas.to_datetime(self._statement['Completed Date'], format='%Y-%m-%d %H:%M:%S')
        self.coinbase_assets = QuoteDownloader.Coinbase_GetCurrencyList()

        self._data[FOF.ACCOUNTS].append({'id': self.ACCOUNT_ID,
                                         'selection_text': self.tr("Revolut statement doesn't have account number.\nPlease select an account for import:")})

        for i, row in self._statement.iterrows():
            assert row['Type'] == 'REWARD', "Buy/Sell operations are not implemented yet"
            assert row['State'] == 'COMPLETED', "States other then COMPLETED were not tested"
            qty = row['Amount']
            amount = row['Fiat amount']
            op_date = row['Completed Date']
            currency_id = self.currency_id(row['Base currency'])
            asset_id = self.asset_id({'type': FOF.ASSET_CRYPTO, 'symbol': row['Currency'],
                           'name': row['Currency'], 'currency': currency_id, 'search_online': 'Coinbase'})
            balance = row['Balance']  # end crypto balance after operation
            vesting = {'id': i, 'type': FOF.PAYMENT_STOCK_VESTING, 'account': self.ACCOUNT_ID, 'timestamp': int(op_date.timestamp()),
                       'asset': asset_id, 'amount': qty, 'price': amount/qty, 'description': row['Description']}
            self._data[FOF.ASSET_PAYMENTS].append(vesting)
        self.assign_symbol_names()
        logging.info(self.tr("Statement loaded successfully: ") + f"{filename}")

    def assign_symbol_names(self):
        for symbol in self._data[FOF.SYMBOLS]:
            asset = self._find_in_list(self._data[FOF.ASSETS], 'id', symbol['asset'])
            if asset['type'] != FOF.ASSET_CRYPTO:
                continue
            coinbase_asset = self._find_in_list(self.coinbase_assets, 'symbol', symbol['symbol'])
            if coinbase_asset is None:
                continue
            asset['name'] = coinbase_asset['name']
            symbol['note'] = 'COIN'



from decimal import Decimal
from functools import partial
from PySide6.QtCore import Qt, Slot, QObject, QDateTime, QAbstractTableModel
from PySide6.QtGui import QFont
from jal.reports.reports import Reports
from jal.db.deposit import JalDeposit
from jal.ui.reports.ui_term_deposits_report import Ui_TermDepositsReportWidget
from jal.widgets.delegates import FloatDelegate, TimestampDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "TermDepositsReport"


# ----------------------------------------------------------------------------------------------------------------------
class DepositsListModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [self.tr("Name"), self.tr("Start Date"), self.tr("End Date"), self.tr("Currency"),
                         self.tr("Initial amount"), self.tr("Accrued Interest"), self.tr("Planned End Value")]
        self._timestamp = 0
        self._view = parent_view
        self._data = []
        self._float_delegate = None
        self._timestamp_delegate = None
        self._initial_total = Decimal('0')
        self._accrued_total = Decimal('0')
        self._planned_total = Decimal('0')

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def headerWidth(self, section):
        return self._view.horizontalHeader().sectionSize(section)

    def data(self, index, role=Qt.DisplayRole, field=''):
        if role == Qt.DisplayRole:
            deposit = self._data[index.row()]
            return self.data_text(deposit, index.column())

    def data_text(self, deposit: JalDeposit, column: int):
        if column == 0:
            return deposit.name()
        if column == 1:
            return deposit.start_date()
        if column == 2:
            return deposit.end_date()
        if column == 3:
            return deposit.currency().symbol()
        if column == 4:
            return deposit.open_amount()
        if column == 5:
            return deposit.accrued_interest(self._timestamp)
        if column == 6:
            return deposit.close_amount()
        return ''

    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section ==0:
                return self.tr("Total")
            if section == 4:
                return self._initial_total
            if section == 5:
                return self._accrued_total
            if section == 6:
                return self._planned_total
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section > 3:
                return Qt.AlignRight | Qt.AlignVCenter
            else:
                return Qt.AlignLeft | Qt.AlignVCenter
        return None

    def updateView(self, timestamp):
        if self._timestamp != timestamp:
            self._timestamp = timestamp
            self.prepareData()
            self.configureView()

    def prepareData(self):
        self._data = JalDeposit.get_term_deposits(self._timestamp)
        self._initial_total = sum([x.open_amount() for x in self._data])
        self._accrued_total = sum([x.accrued_interest(self._timestamp) for x in self._data])
        self._planned_total = sum([x.close_amount() for x in self._data])
        self.modelReset.emit()

    def configureView(self):
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._float_delegate = FloatDelegate(2, allow_tail=False)
        self._timestamp_delegate = TimestampDelegate(display_format='%d/%m/%Y')
        self._view.setItemDelegateForColumn(1, self._timestamp_delegate)
        self._view.setItemDelegateForColumn(2, self._timestamp_delegate)
        self._view.setItemDelegateForColumn(4, self._float_delegate)
        self._view.setItemDelegateForColumn(5, self._float_delegate)
        self._view.setItemDelegateForColumn(6, self._float_delegate)
        self._view.setColumnWidth(0, 200)
        self._view.setColumnWidth(1, self._view.fontMetrics().horizontalAdvance("00/00/0000") * 1.1)
        self._view.setColumnWidth(2, self._view.fontMetrics().horizontalAdvance("00/00/0000") * 1.1)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 150)
        self._view.setColumnWidth(5, 150)
        self._view.setColumnWidth(6, 150)

# ----------------------------------------------------------------------------------------------------------------------
class TermDepositsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Term deposits")
        self.window_class = "TermDepositsReportWindow"

# ----------------------------------------------------------------------------------------------------------------------
class TermDepositsReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_TermDepositsReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Term deposits")

        self.payments_model = DepositsListModel(self.ui.ReportTableView)
        self.ui.ReportTableView.setModel(self.payments_model)

        self.connect_signals_and_slots()

        current_time = QDateTime.currentDateTime()
        current_time.setTimeSpec(Qt.UTC)  # We use UTC everywhere so need to force TZ info
        self.ui.DepositsDate.setDateTime(current_time)

    def connect_signals_and_slots(self):
        self.ui.DepositsDate.dateChanged.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTableView.model()))

    @Slot()
    def updateReport(self):
        self.ui.ReportTableView.model().updateView(timestamp=self.ui.DepositsDate.date().endOfDay(Qt.UTC).toSecsSinceEpoch())



from functools import partial
from PySide6.QtCore import Qt, Slot, QObject, QDateTime, QMargins
from PySide6.QtWidgets import QWidget, QHBoxLayout
from PySide6.QtCharts import QChartView, QLineSeries, QDateTimeAxis, QValueAxis
from jal.reports.reports import Reports
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.ui.reports.ui_account_balance_report import Ui_AccountBalanceHistoryReportWidget
from jal.widgets.mdi import MdiWidget
from jal.widgets.helpers import timestamp_range

JAL_REPORT_CLASS = "AccountBalanceHistoryReport"

# ----------------------------------------------------------------------------------------------------------------------
class BalanceChartWidget(QWidget):
    def __init__(self, parent): #, quotes, trades, data_range, currency_name):
        super().__init__(parent=parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)
        self.chartView = QChartView()
        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove extra space around layout
        self.layout.addWidget(self.chartView)
        self.setLayout(self.layout)
        self.axisX = None
        self.axisY = None

    def updateView(self, balances, currency_name):
        self.balances_series = QLineSeries()
        for point in balances:            # Conversion to 'float' in order not to get 'int' overflow on some platforms
            self.balances_series.append(float(point['timestamp']), point['balance'])

        # Cleanup if anything already drawn in the chart
        self.chartView.chart().removeAllSeries()
        axes = self.chartView.chart().axes()
        for axis in axes:
            self.chartView.chart().removeAxis(axis)

        # Create new X-axis
        self.axisX = QDateTimeAxis()
        self.axisX.setFormat("yyyy/MM/dd")
        self.axisX.setLabelsAngle(-90)
        self.axisX.setTitleText("Date")
        min_ts = int(min([x['timestamp'] for x in balances]) / 1000)
        max_ts = int(max([x['timestamp'] for x in balances]) / 1000)
        self.axisX.setRange(QDateTime().fromSecsSinceEpoch(min_ts), QDateTime().fromSecsSinceEpoch(max_ts))

        # Create new Y-axis
        self.axisY = QValueAxis()
        self.axisY.setTitleText("Balance, " + currency_name)

        # Arrange chart
        self.chartView.chart().addSeries(self.balances_series)
        self.chartView.chart().addAxis(self.axisX, Qt.AlignBottom)
        self.balances_series.attachAxis(self.axisX)
        self.chartView.chart().addAxis(self.axisY, Qt.AlignLeft)
        self.balances_series.attachAxis(self.axisY)
        self.chartView.chart().legend().hide()
        self.chartView.setViewportMargins(0, 0, 0, 0)
        self.chartView.chart().layout().setContentsMargins(0, 0, 0, 0)  # To remove extra spacing around chart
        self.chartView.chart().setBackgroundRoundness(0)  # To remove corner rounding
        self.chartView.chart().setMargins(QMargins(0, 0, 0, 0))  # Allow chart to fill all space

# ----------------------------------------------------------------------------------------------------------------------
class AccountBalanceHistoryReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Account balance history")
        self.window_class = "AccountBalanceHistoryReportWindow"

# ----------------------------------------------------------------------------------------------------------------------
class AccountBalanceHistoryReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_AccountBalanceHistoryReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Account balance")

        self.chart = BalanceChartWidget(self)
        self.ui.reportLayout.addWidget(self.chart)

        if settings is not None:
            self.ui.ReportRange.setRange(settings['begin_ts'], settings['end_ts'])
            self.ui.ReportAccountButton.account_id = settings['account_id']
        self.connect_signals_and_slots()
        self.updateReport()

    def connect_signals_and_slots(self):
        self.ui.ReportRange.changed.connect(self.updateReport)
        self.ui.ReportAccountButton.changed.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_chart, self.chart.chartView))

    @Slot()
    def updateReport(self):
        if self.ui.ReportAccountButton.account_id == 0:
            return
        account = JalAccount(self.ui.ReportAccountButton.account_id)
        balances = []
        date_range = self.ui.ReportRange.getRange()
        for ts in timestamp_range(date_range[0], date_range[1]):
            balances.append({'timestamp': ts*1000, 'balance': account.balance(ts)})
        self.chart.updateView(balances, JalAsset(account.currency()).symbol())



import logging
from decimal import Decimal
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QMessageBox, QHeaderView
from PySide6.QtSql import QSqlTableModel
from PySide6.QtGui import QFont
from jal.ui.widgets.ui_term_deposit_operation import Ui_TermDepositOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.constants import DepositActions
from jal.db.view_model import JalViewModel
from jal.db.operations import LedgerTransaction
from jal.db.helpers import load_icon, now_ts, localize_decimal, db_row2dict
from jal.widgets.delegates import FloatDelegate, TimestampDelegate, ConstantLookupDelegate


# ----------------------------------------------------------------------------------------------------------------------
class TermDepositWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_TermDepositOperation)
        self.operation_type = LedgerTransaction.TermDeposit
        super()._init_db("term_deposits")

        self.timestamp_delegate = TimestampDelegate()
        self.float_delegate = FloatDelegate(2)
        self.action_delegate = ConstantLookupDelegate(DepositActions, self)

        self.ui.actions_table.horizontalHeader().setFont(self.bold_font)
        self.ui.add_button.setIcon(load_icon("add.png"))
        self.ui.del_button.setIcon(load_icon("remove.png"))

        self.ui.add_button.clicked.connect(self.add_action)
        self.ui.del_button.clicked.connect(self.remove_action)

        self.actions_model = DepositActionsModel(self.ui.actions_table, "deposit_actions")
        self.actions_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.ui.actions_table.setModel(self.actions_model)
        self.actions_model.dataChanged.connect(self.onDataChange)
        self.ui.account_widget.changed.connect(self.mapper.submit)

        self.mapper.addMapping(self.ui.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.ui.actions_table.setItemDelegateForColumn(2, self.timestamp_delegate)
        self.ui.actions_table.setItemDelegateForColumn(3, self.action_delegate)
        self.ui.actions_table.setItemDelegateForColumn(4, self.float_delegate)

        self.model.select()
        self.actions_model.select()
        self.actions_model.configureView()

    def set_id(self, oid):
        self.actions_model.setFilter(f"deposit_actions.deposit_id = {oid}")  # First we need to select right children
        super().set_id(oid)

    @Slot()
    def add_action(self):
        new_record = self.actions_model.record()
        new_record.setValue("timestamp", now_ts())
        new_record.setValue("action_type", 0)
        new_record.setValue("amount", '0')
        if not self.actions_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.actions_model.lastError().text())
            return

    @Slot()
    def remove_action(self):
        selection = self.ui.actions_table.selectionModel().selection().indexes()
        for idx in selection:
            self.actions_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    def _validated(self):
        open_action_count = 0
        close_action_count = 0
        fields = db_row2dict(self.model, 0)
        if not fields['account_id']:
            return False
        if not self.actions_model.rowCount():
            QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Deposit contains no actions"), QMessageBox.Ok)
            return False
        for row in range(self.actions_model.rowCount()):
            if self.actions_model.row_is_deleted(row):
                continue
            fields = db_row2dict(self.actions_model, row)
            if fields['action_type'] is None or fields['action_type'] == 0:
                QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Deposit action type isn't set"), QMessageBox.Ok)
                return False
            if fields['action_type'] == DepositActions.Opening:
                open_action_count += 1
            if fields['action_type'] == DepositActions.Closing:
                close_action_count += 1
        if open_action_count != 1:
            QMessageBox().warning(self, self.tr("Incorrect data"), self.tr("There should be 1 deposit opening action"), QMessageBox.Ok)
            return False
        if close_action_count != 1:
            QMessageBox().warning(self, self.tr("Incorrect data"), self.tr("There should be 1 deposit closing action"), QMessageBox.Ok)
            return False
        return True

    def _save(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            oid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if oid is None:  # we just have saved new action record and need last inserted id
                oid = self.model.last_insert_id()
            for row in range(self.actions_model.rowCount()):   # Set PID for all child records
                self.actions_model.setData(self.actions_model.index(row, self.actions_model.fieldIndex("deposit_id")), oid)
            if not self.actions_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.actions_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.model.database().commit()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.actions_model.revertAll()
        self.modified = False
        self.ui.commit_button.setEnabled(False)
        self.ui.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.actions_model.setFilter(f"deposit_actions.deposit_id = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("account_id", account_id)
        new_record.setValue("note", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.actions_model.rowCount()):
            child_records.append(self.actions_model.record(row))
        self.actions_model.setFilter(f"deposit_actions.deposit_id = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("deposit_id")
            assert self.actions_model.insertRows(0, 1)
            self.actions_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        return new_record


# ----------------------------------------------------------------------------------------------------------------------
class DepositActionsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "deposit_id", self.tr("Date/Time"), self.tr("Action"), self.tr("Amount")]

    def configureView(self):
        assert self.columnCount() == len(self._columns), "Mismatch between deposit actions model and view"
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnWidth(2, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self._view.setColumnWidth(4, 100)
        self._view.horizontalHeader().setSectionResizeMode(3, QHeaderView.Stretch)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'peer_report.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QFrame, QGridLayout,
    QHeaderView, QLabel, QSizePolicy, QSpacerItem,
    QSplitter, QVBoxLayout, QWidget)

from jal.widgets.account_select import CurrencyComboBox
from jal.widgets.custom.date_range_selector import DateRangeSelector
from jal.widgets.custom.tableview_with_footer import TableViewWithFooter
from jal.widgets.operations_tabs import JalOperationsTabs
from jal.widgets.reference_selector import PeerSelector

class Ui_PeerReportWidget(object):
    def setupUi(self, PeerReportWidget):
        if not PeerReportWidget.objectName():
            PeerReportWidget.setObjectName(u"PeerReportWidget")
        PeerReportWidget.resize(767, 345)
        self.verticalLayout = QVBoxLayout(PeerReportWidget)
        self.verticalLayout.setSpacing(2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.ReportParamsFrame = QFrame(PeerReportWidget)
        self.ReportParamsFrame.setObjectName(u"ReportParamsFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ReportParamsFrame.sizePolicy().hasHeightForWidth())
        self.ReportParamsFrame.setSizePolicy(sizePolicy)
        self.ReportParamsFrame.setFrameShape(QFrame.Panel)
        self.ReportParamsFrame.setFrameShadow(QFrame.Sunken)
        self.gridLayout = QGridLayout(self.ReportParamsFrame)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(2, 2, 2, 2)
        self.ReportPeerEdit = PeerSelector(self.ReportParamsFrame)
        self.ReportPeerEdit.setObjectName(u"ReportPeerEdit")

        self.gridLayout.addWidget(self.ReportPeerEdit, 0, 2, 1, 1)

        self.ReportFrameSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.ReportFrameSpacer, 0, 5, 1, 1)

        self.ReportRange = DateRangeSelector(self.ReportParamsFrame)
        self.ReportRange.setObjectName(u"ReportRange")
        self.ReportRange.setProperty("ItemsList", u"QTD;YTD;this_year;last_year")

        self.gridLayout.addWidget(self.ReportRange, 0, 0, 1, 1)

        self.ReportPeerLbl = QLabel(self.ReportParamsFrame)
        self.ReportPeerLbl.setObjectName(u"ReportPeerLbl")

        self.gridLayout.addWidget(self.ReportPeerLbl, 0, 1, 1, 1)

        self.TotalCurrencyLbl = QLabel(self.ReportParamsFrame)
        self.TotalCurrencyLbl.setObjectName(u"TotalCurrencyLbl")

        self.gridLayout.addWidget(self.TotalCurrencyLbl, 0, 3, 1, 1)

        self.TotalCurrencyCombo = CurrencyComboBox(self.ReportParamsFrame)
        self.TotalCurrencyCombo.setObjectName(u"TotalCurrencyCombo")

        self.gridLayout.addWidget(self.TotalCurrencyCombo, 0, 4, 1, 1)


        self.verticalLayout.addWidget(self.ReportParamsFrame)

        self.splitter = QSplitter(PeerReportWidget)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Vertical)
        self.ReportTableView = TableViewWithFooter(self.splitter)
        self.ReportTableView.setObjectName(u"ReportTableView")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(4)
        sizePolicy1.setHeightForWidth(self.ReportTableView.sizePolicy().hasHeightForWidth())
        self.ReportTableView.setSizePolicy(sizePolicy1)
        self.ReportTableView.setFrameShape(QFrame.Panel)
        self.ReportTableView.setFrameShadow(QFrame.Sunken)
        self.ReportTableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.ReportTableView.setAlternatingRowColors(True)
        self.ReportTableView.setGridStyle(Qt.DotLine)
        self.ReportTableView.setWordWrap(False)
        self.splitter.addWidget(self.ReportTableView)
        self.ReportTableView.verticalHeader().setVisible(False)
        self.ReportTableView.verticalHeader().setMinimumSectionSize(20)
        self.ReportTableView.verticalHeader().setDefaultSectionSize(20)
        self.OperationDetails = JalOperationsTabs(self.splitter)
        self.OperationDetails.setObjectName(u"OperationDetails")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(1)
        sizePolicy2.setHeightForWidth(self.OperationDetails.sizePolicy().hasHeightForWidth())
        self.OperationDetails.setSizePolicy(sizePolicy2)
        self.splitter.addWidget(self.OperationDetails)

        self.verticalLayout.addWidget(self.splitter)


        self.retranslateUi(PeerReportWidget)

        QMetaObject.connectSlotsByName(PeerReportWidget)
    # setupUi

    def retranslateUi(self, PeerReportWidget):
        PeerReportWidget.setWindowTitle(QCoreApplication.translate("PeerReportWidget", u"Report by peer", None))
        self.ReportPeerLbl.setText(QCoreApplication.translate("PeerReportWidget", u"Peer:", None))
        self.TotalCurrencyLbl.setText(QCoreApplication.translate("PeerReportWidget", u"Common currency:", None))
    # retranslateUi

